/* Copyright (c) 2012-2017 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "MainWindow.h"
#include "Database.h"
#include "AboutWindow.h"
#include "SettingsPage.h"
#include "PlayersPage.h"
#include "ScorePage.h"
#include "RankingPage.h"
#include "ExportDialog.h"

// Qt includes
#include <QLabel>
#include <QVBoxLayout>
#include <QFont>
#include <QTabWidget>
#include <QMenuBar>
#include <QAction>
#include <QApplication>
#include <QStatusBar>
#include <QMessageBox>
#include <QUrl>
#include <QDesktopServices>
#include <QFileInfo>
#include <QDebug>
#include <QFileDialog>
#include <QDir>
#include <QStandardPaths>
#include <QCoreApplication>
#include <QDirIterator>

MainWindow::MainWindow()
{
    // For the static Windows build, we embed the icon. So let's see if we have it
    QDirIterator it(QString::fromUtf8(":"), QDirIterator::Subdirectories);
    while (it.hasNext()) {
        if (it.next() == QString::fromUtf8(":/icon.png")) {
            m_iconPath = QString::fromUtf8(":/icon.png");
            break;
        }
    }
    if (m_iconPath.isEmpty()) {
        m_iconPath = absolutePath(QString::fromUtf8("icon.png"));
    }

    // Setup the database
    m_db = new Database(this);

    // Window setup
    setWindowIcon(QIcon(m_iconPath));
    setWindowTitle(tr("Muckturnier"));

    QWidget *mainWidget = new QWidget(this);
    QVBoxLayout *layout = new QVBoxLayout(mainWidget);
    setCentralWidget(mainWidget);

    m_headerLabel = new QLabel;
    m_headerLabel->setAlignment(Qt::AlignHCenter);
    QFont headerFont = m_headerLabel->font();
    headerFont.setPointSize(headerFont.pointSize() * 1.5);
    headerFont.setWeight(QFont::Bold);
    m_headerLabel->setFont(headerFont);
    layout->addWidget(m_headerLabel);

    m_tabWidget = new QTabWidget;
    layout->addWidget(m_tabWidget);
    connect(m_tabWidget, &QTabWidget::currentChanged, this, &MainWindow::tabChanged);
    m_tabWidget->setEnabled(false);

    // Tab setup

    m_tabWidget->blockSignals(true);

    m_settingsPage = new SettingsPage(this, m_db);
    m_tabWidget->addTab(m_settingsPage, tr("Einstellungen"));
    connect(m_settingsPage, &SettingsPage::statusUpdate, this, &MainWindow::updateStatus);
    connect(m_settingsPage, &SettingsPage::settingsSaved, this, &MainWindow::settingsSaved);
    connect(m_settingsPage, &SettingsPage::dbReset, this, &MainWindow::dbReset);

    m_playersPage = new PlayersPage(this, m_db);
    m_tabWidget->addTab(m_playersPage, tr("Paare anmelden"));
    m_tabWidget->setTabEnabled(m_tabWidget->indexOf(m_playersPage), false);
    connect(m_playersPage, &PlayersPage::statusUpdate, this, &MainWindow::updateStatus);
    connect(m_playersPage, &PlayersPage::tournamentCanBeStarted,
            this, &MainWindow::setScoreTabEnabled);

    m_scorePage = new ScorePage(this, m_db);
    m_tabWidget->addTab(m_scorePage, tr("Ergebnisse eintragen"));
    m_tabWidget->setTabEnabled(m_tabWidget->indexOf(m_scorePage), false);
    connect(m_scorePage, &ScorePage::tournamentStarted, this, &MainWindow::tournamentStarted);
    connect(m_scorePage, &ScorePage::statusUpdate, this, &MainWindow::updateStatus);

    m_rankingPage = new RankingPage(this, m_db);
    m_tabWidget->addTab(m_rankingPage, tr("Rangliste"));
    m_tabWidget->setTabEnabled(m_tabWidget->indexOf(m_rankingPage), false);
    connect(m_rankingPage, &RankingPage::statusUpdate, this, &MainWindow::updateStatus);
    connect(m_playersPage, &PlayersPage::playersListChanged, m_rankingPage, &RankingPage::reload);
    connect(m_scorePage, &ScorePage::scoreChanged, m_rankingPage, &RankingPage::reload);

    m_tabWidget->blockSignals(false);

    // Menu

    QMenu *fileMenu = menuBar()->addMenu(tr("Datei"));

    QAction *newTournamentAction = new QAction(tr("Neues Turnier starten"), this);
    fileMenu->addAction(newTournamentAction);
    connect(newTournamentAction, &QAction::triggered, this, &MainWindow::startNewTournament);

    QAction *loadTournamentAction = new QAction(tr("Turnier öffnen"), this);
    fileMenu->addAction(loadTournamentAction);
    connect(loadTournamentAction, &QAction::triggered, this, &MainWindow::triggerLoadTournament);

    fileMenu->addSeparator();

    m_exportAction = new QAction(tr("Turnierdaten exportieren"), this);
    connect(m_rankingPage, &RankingPage::enableExport, m_exportAction, &QAction::setEnabled);
    connect(m_exportAction, &QAction::triggered, this, &MainWindow::showExportDialog);
    fileMenu->addAction(m_exportAction);
    m_exportAction->setEnabled(false);

    fileMenu->addSeparator();

    QAction *quitAction = new QAction(tr("Beenden"), this);
    fileMenu->addAction(quitAction);
    connect(quitAction, &QAction::triggered, this, &MainWindow::quit);

    QMenu *helpMenu = menuBar()->addMenu(tr("Hilfe"));

    QAction *handbookAction = new QAction(tr("Handbuch zu Muckturnier"), this);
    helpMenu->addAction(handbookAction);
    connect(handbookAction, &QAction::triggered, this, &MainWindow::showHandbook);

    QAction *aboutAction = new QAction(tr("Über Muckturnier"), this);
    helpMenu->addAction(aboutAction);
    connect(aboutAction, &QAction::triggered, this, &MainWindow::showAboutWindow);

    updateStatus(tr("Bitte ein neues Turnier starten oder ein bestehendes öffnen!"));
}

void MainWindow::tabChanged(int tab)
{
    // Updates the global header, according to the currently selected tab.

    m_headerLabel->setText(m_tabWidget->tabText(tab).replace(QString::fromUtf8("&"),
                                                             QString::fromUtf8("")));

    if (m_tabWidget->currentIndex() == m_tabWidget->indexOf(m_playersPage)) {
        m_playersPage->setPlayersNameFocus();
    }
}

void MainWindow::startNewTournament()
{
    QString dbFileName = QFileDialog::getSaveFileName(this,
        tr("Bitte einen Dateinamen für die Turnierdatenbank auswählen"),
        QDir::toNativeSeparators(QStandardPaths::writableLocation(QStandardPaths::DocumentsLocation)
                                 + QString::fromUtf8("/Muckturnier.mtdb")),
        tr("Muckturnier-Datenbanken") + QString::fromUtf8(" (*.mtdb);;")
        + tr("Alle Dateien") + QString::fromUtf8(" (*)"));

    if (dbFileName.isEmpty()) {
        return;
    }

    // Yes, we have to do it in this way, because QFileInfo::isWritable
    // does not work as expected on Windows (returns false for "My Documents").
    bool canWrite;
    QFile test(dbFileName);
    if (test.open(QIODevice::WriteOnly)) {
        canWrite = true;
        test.close();
    } else {
        canWrite = false;
    }

    if (! canWrite) {
        QMessageBox::critical(this,
            tr("Neues Turnier starten"),
            tr("Es bestehen keine Schreibrechte für die Datei „%1“. "
               "Bitte die Zugriffsrechte überprüfen!").arg(QDir::toNativeSeparators(dbFileName)),
            QMessageBox::Ok);
        return;
    }

    QFileInfo dbFileInfo = QFileInfo(dbFileName);
    if (dbFileInfo.exists()) {
        // When we're overwriting an existing database, it has to be deleted
        dbFileInfo.dir().remove(dbFileInfo.fileName());
    }

    updateStatus(tr("Lege neue Turnierdatenbank an …"));
    QApplication::setOverrideCursor(Qt::WaitCursor);

    if (! m_db->create(dbFileName)) {
        // This should not happen
        updateStatus(tr("Das Anlegen der Turnierdatenbank „%1“ ist fehlgeschlagen!").arg(
            QDir::toNativeSeparators(dbFileName)));
        QApplication::restoreOverrideCursor();
        return;
    }

    databaseChanged(dbFileName);
    dbReset();
    m_settingsPage->resetGui();
    m_tabWidget->setCurrentIndex(m_tabWidget->indexOf(m_settingsPage));

    updateStatus(tr("Turnier „%1“ gestartet.").arg(QDir::toNativeSeparators(dbFileName)));
    QApplication::restoreOverrideCursor();
}

void MainWindow::triggerLoadTournament()
{
    loadTournament();
}

void MainWindow::loadTournament(QString dbToOpen)
{
    QString dbFileName;
    if (! dbToOpen.isEmpty()) {
        dbFileName = dbToOpen;
    } else {
        dbFileName = QFileDialog::getOpenFileName(this,
            tr("Bitte eine Turnierdatenbank auswählen"),
            QDir::toNativeSeparators(
                QStandardPaths::writableLocation(QStandardPaths::DocumentsLocation)),
            tr("Muckturnier-Datenbanken") + QString::fromUtf8(" (*.mtdb *.db);;")
            + tr("Alle Dateien") + QString::fromUtf8(" (*)"));
        if (dbFileName.isEmpty()) {
            return;
        }
    }

    QFileInfo fileInfo = QFileInfo(dbFileName);

    if (! fileInfo.exists()) {
        QMessageBox::critical(this,
            tr("Turnier öffnen"),
            tr("Das Öffnen der Turnierdatenbank „%1“ ist nicht möglich:<br/>"
               "Die Datei konnte nicht gefunden werden.").arg(
               QDir::toNativeSeparators(dbFileName)),
            QMessageBox::Ok);
        return;
    }

    if (! (fileInfo.isFile() || fileInfo.isSymLink())) {
        QMessageBox::critical(this,
            tr("Turnier öffnen"),
            tr("Das Öffnen der Turnierdatenbank „%1“ ist nicht möglich:<br/>"
               "Es ist weder eine Datei, noch ein symbolischer Link.").arg(
               QDir::toNativeSeparators(dbFileName)),
            QMessageBox::Ok);
        return;
    }

    if (! fileInfo.isReadable()) {
        QMessageBox::critical(this,
            tr("Turnier öffnen"),
            tr("Das Öffnen der Turnierdatenbank „%1“ ist nicht möglich:<br/>"
               "Die Datei ist nicht lesbar. "
               "Bitte die Zugriffsrechte überprüfen!").arg(
               QDir::toNativeSeparators(dbFileName)),
            QMessageBox::Ok);
        return;
    }

    if (! fileInfo.isWritable()) {
        QMessageBox::critical(this,
            tr("Turnier öffnen"),
            tr("Das Laden der Turnierdatenbank „%1“ ist nicht möglich:<br/>"
               "Die Datei ist nicht beschreibbar. "
               "Bitte die Zugriffsrechte überprüfen!").arg(
               QDir::toNativeSeparators(dbFileName)),
            QMessageBox::Ok);
        return;
    }

    // At least, we can read and write the file. So let's try to open it

    updateStatus(tr("Öffne Turnierdatenbank …"));
    QApplication::setOverrideCursor(Qt::WaitCursor);

    if (! m_db->openDb(dbFileName)) {
        QApplication::restoreOverrideCursor();
        QMessageBox::critical(this,
            tr("Turnier öffnen"),
            tr("<p>Das Öffnen der Turnierdatenbank „%1“ ist nicht möglich:</p><p>%2</p>").arg(
                QDir::toNativeSeparators(dbFileName), m_db->openError()),
            QMessageBox::Ok);
        return;
    }

    databaseChanged(dbFileName);

    if (m_db->tournamentStarted()) {
        m_settingsPage->setToSaved();
        m_tabWidget->setCurrentIndex(m_tabWidget->indexOf(m_scorePage));
    } else {
        // The score page could be enabled by the previously used database,
        // so disable it first
        m_tabWidget->setTabEnabled(m_tabWidget->indexOf(m_scorePage), false);
        // Re-enable the score page if the database already contains
        // the necessary number of players/pairs
        m_playersPage->reload();

        if (m_db->playersCount() == 0) {
            // A completely empty database has been loaded
            m_settingsPage->resetGui();
            m_tabWidget->setTabEnabled(m_tabWidget->indexOf(m_playersPage), false);
            m_tabWidget->setCurrentIndex(m_tabWidget->indexOf(m_settingsPage));
        } else {
            m_settingsPage->setToSaved();
            m_tabWidget->setCurrentIndex(m_tabWidget->indexOf(m_playersPage));
        }
    }

    updateStatus(tr("Turnierdatenbank %1 geöffnet").arg(QDir::toNativeSeparators(dbFileName)));

    QApplication::restoreOverrideCursor();
}

void MainWindow::quit()
{
    QApplication::quit();
}

void MainWindow::showHandbook()
{
    // Opens the default browser to display the handbook.

    QString readmePath = absolutePath(QString::fromUtf8("Readme.htm"));
    QFileInfo fileInfo(readmePath);
    if (fileInfo.exists() && (fileInfo.isFile() || fileInfo.isSymLink())) {
        QDesktopServices::openUrl(QUrl::fromLocalFile(readmePath));
        return;
    }

    QMessageBox::critical(this,
                          tr("Handbuch anzeigen"),
                          tr("Das Handbuch kann nicht angezeigt werden: "
                             "Die Datei „Readme.htm“ wurde nicht gefunden."),
                          QMessageBox::Ok);
}

void MainWindow::showAboutWindow()
{
    // Shows the 'about' window.
    QDialog *aboutWindow = new AboutWindow(this, m_iconPath);
    aboutWindow->show();
}

void MainWindow::updateStatus(QString message)
{
    // Displays a message in the status bar for 5 seconds.
    statusBar()->showMessage(message, 5000);
}

void MainWindow::databaseChanged(QString dbFileName)
{
    // Updates the GUI after loading or creating a database.

    setWindowTitle(tr("Muckturnier: %1").arg(QFileInfo(dbFileName).fileName()));
    m_tabWidget->setEnabled(true);

    m_settingsPage->reload();

    setPlayersTabName();
    m_tabWidget->setTabEnabled(m_tabWidget->indexOf(m_playersPage), true);
    m_playersPage->reload();

    m_exportAction->setEnabled(false);

    checkEnableRankingPage();
    if (m_db->tournamentStarted()) {
        m_rankingPage->reload();
    }
}

void MainWindow::dbReset()
{
    setPlayersTabName();
    m_tabWidget->setTabEnabled(m_tabWidget->indexOf(m_playersPage), false);
    m_playersPage->reset();
    m_tabWidget->setTabEnabled(m_tabWidget->indexOf(m_scorePage), false);
    m_tabWidget->setTabEnabled(m_tabWidget->indexOf(m_rankingPage), false);
}

void MainWindow::settingsSaved()
{
    setPlayersTabName();
    m_tabWidget->setTabEnabled(m_tabWidget->indexOf(m_playersPage), true);
    m_playersPage->reload();
    m_tabWidget->setCurrentIndex(m_tabWidget->indexOf(m_playersPage));
}

void MainWindow::setScoreTabEnabled(bool state)
{
    m_tabWidget->setTabEnabled(m_tabWidget->indexOf(m_scorePage), state);
    if(state == true) {
        m_scorePage->reload();
    }
}

void MainWindow::tournamentStarted()
{
    setPlayersTabName();
    m_playersPage->reload();
    checkEnableRankingPage();
}

void MainWindow::setPlayersTabName()
{
    if (m_db->tournamentStarted()) {
        m_tabWidget->setTabText(m_tabWidget->indexOf(m_playersPage),
                                tr("Angemeldete %1").arg(
                                   m_db->playersString(Database::Numerus::plural)));
    } else {
        m_tabWidget->setTabText(m_tabWidget->indexOf(m_playersPage),
                                tr("%1 anmelden").arg(
                                   m_db->playersString(Database::Numerus::plural)));
    }
}

void MainWindow::checkEnableRankingPage()
{
    m_tabWidget->setTabEnabled(m_tabWidget->indexOf(m_rankingPage),
                               m_db->tournamentStarted());
}

QString MainWindow::absolutePath(QString fileName)
{
    return QDir::toNativeSeparators(QCoreApplication::applicationDirPath())
           + QDir::separator()
           + fileName;
}

void MainWindow::resizeEvent(QResizeEvent *event)
{
   QMainWindow::resizeEvent(event);
   m_scorePage->setScrollAreaHeight();
}

void MainWindow::showExportDialog()
{
    ExportDialog *exportDialog = new ExportDialog(this, m_db, m_rankingPage);
    connect(exportDialog, &ExportDialog::statusUpdate, this, &MainWindow::updateStatus);
    exportDialog->exec();
}
