/* Copyright (c) 2012-2017 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "PlayersPage.h"
#include "Database.h"
#include "Cell.h"

// Qt includes
#include <QDebug>
#include <QVBoxLayout>
#include <QGroupBox>
#include <QHBoxLayout>
#include <QLabel>
#include <QLineEdit>
#include <QPushButton>
#include <QTableWidget>
#include <QHeaderView>
#include <QScrollBar>
#include <QMenu>
#include <QAction>
#include <QMessageBox>
#include <QApplication>
#include <QWidgetAction>

PlayersPage::PlayersPage(QWidget *parent, Database *db)
    : QWidget(parent), m_db(db)
{
    QVBoxLayout *layout = new QVBoxLayout(this);

    // Register new players/pairs

    m_newPlayersBox = new QGroupBox;
    QHBoxLayout *newPlayersBoxLayout = new QHBoxLayout(m_newPlayersBox);
    layout->addWidget(m_newPlayersBox);

    m_nameLabel = new QLabel;
    m_nameLabel->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    newPlayersBoxLayout->addWidget(m_nameLabel);

    m_playersName = new QLineEdit;
    newPlayersBoxLayout->addWidget(m_playersName);
    connect(m_playersName, &QLineEdit::textChanged, this, &PlayersPage::playersNameChanged);
    connect(m_playersName, &QLineEdit::returnPressed, this, &PlayersPage::registerPlayers);

    m_registerButton = new QPushButton(tr("Anmelden"));
    m_registerButton->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    m_registerButton->setEnabled(false);
    newPlayersBoxLayout->addWidget(m_registerButton);
    connect(m_registerButton, &QPushButton::clicked, this, &PlayersPage::registerPlayers);

    // Registered players/pairs

    m_playersBox = new QGroupBox;
    QVBoxLayout *playersBoxLayout = new QVBoxLayout(m_playersBox);
    layout->addWidget(m_playersBox);

    m_playersList = new QTableWidget(0, 1);
    m_playersList->horizontalHeader()->hide();
    m_playersList->verticalHeader()->hide();
    m_playersList->setShowGrid(false);
    m_playersList->setWordWrap(false);
    m_playersList->setSelectionMode(QTableWidget::NoSelection);
    connect(m_playersList, &QTableWidget::cellDoubleClicked, this, &PlayersPage::cacheOriginalName);
    connect(m_playersList, &QTableWidget::cellChanged, this, &PlayersPage::editEntered);
    connect(m_playersList->verticalScrollBar(), &QScrollBar::valueChanged,
            m_playersList, &QTableWidget::resizeColumnsToContents);
    m_playersList->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(m_playersList, &QTableWidget::customContextMenuRequested,
            this, &PlayersPage::showContextMenu);

    playersBoxLayout->addWidget(m_playersList);

    m_stateLabel = new QLabel;
    layout->addWidget(m_stateLabel);

    // Context menu

    m_contextMenu = new QMenu(this);

    m_menuTitleAction = new QWidgetAction(this);
    m_menuTitle = new QLabel;
    m_menuTitle->setAlignment(Qt::AlignCenter);
    m_menuTitle->setMargin(4);
    m_menuTitleAction->setDefaultWidget(m_menuTitle);

    m_editAction = new QAction(this);
    connect(m_editAction, &QAction::triggered, this, &PlayersPage::editPlayers);

    m_deleteAction = new QAction(this);
    connect(m_deleteAction, &QAction::triggered, this, &PlayersPage::deletePlayers);
}

void PlayersPage::setPlayersNameFocus()
{
    m_playersName->setFocus();
}

void PlayersPage::reload()
{
    m_newPlayersBox->show();
    m_newPlayersBox->setTitle(tr("Neue %1 anmelden").arg(
                                 m_db->playersString(Database::Numerus::plural)));
    m_playersBox->setTitle(tr("Angemeldete %1").arg(
                              m_db->playersString(Database::Numerus::plural)));
    m_deleteAction->setText(tr("%1 löschen").arg(m_db->playersString(Database::Numerus::singular)));
    switch (m_db->getTournamentMode()) {
    case Database::TournamentMode::fixedPairs:
        m_nameLabel->setText(tr("Name(n):"));
        m_editAction->setText(tr("Name(n) ändern"));
        break;
    case Database::TournamentMode::singlePlayers:
        m_nameLabel->setText(tr("Name:"));
        m_editAction->setText(tr("Name ändern"));
        break;
    }

    m_playersName->setFocus();
    m_playersList->clear();

    int playersCount = 0;
    QStringList allPlayers = m_db->listPlayers();
    m_playersList->clear();
    m_playersList->setRowCount(allPlayers.size());
    int rowType = 1;

    m_playersList->blockSignals(true);

    for (const QString &player : allPlayers) {
        rowType++;
        rowType = rowType % 2;
        m_playersList->setItem(playersCount, 0, new Cell(player, true, rowType));
        playersCount++;
    }

    m_playersList->resizeColumnsToContents();

    m_playersList->blockSignals(false);

    QString state;
    if (playersCount == 0) {
        state = tr("Keine %1 angemeldet").arg(m_db->playersString(Database::Numerus::plural));
    } else if (playersCount == 1) {
        state = tr("1 %1 angemeldet").arg(m_db->playersString(Database::Numerus::singular));
    } else {
        state = tr("%1 %2 angemeldet").arg(QString::number(playersCount),
                                           m_db->playersString(Database::Numerus::plural));
    }

    if (playersCount % m_db->necessaryDivisor() != 0) {
        state.append(tr(". Die Anzahl der %1 muss durch %2 teilbar sein, "
                        "um mit dem Turnier beginnen zu können!").arg(
                        m_db->playersString(Database::Numerus::plural),
                        QString::number(m_db->necessaryDivisor())));
        emit tournamentCanBeStarted(false);
    } else {
        if (playersCount > 0) {
            emit tournamentCanBeStarted(true);
        }
    }

    m_stateLabel->setText(state);

    if (m_db->tournamentStarted()) {
        m_newPlayersBox->hide();
    }
}

void PlayersPage::reset()
{
    m_playersList->clear();
    m_newPlayersBox->show();
}

void PlayersPage::playersNameChanged(QString name)
{
    // Checks if a player/pair name is empty and enables/disables the 'register' button
    m_registerButton->setEnabled(! name.simplified().isEmpty());
}

void PlayersPage::registerPlayers()
{
    if (! m_registerButton->isEnabled()) {
        return;
    }

    QString name = m_playersName->text().simplified();

    if (!m_db->checkName(name)) {
        QMessageBox::critical(this,
                              m_newPlayersBox->title(),
                              tr("Der Name „%1“ ist bereits vergeben!").arg(name),
                              QMessageBox::Ok);
        return;
    }

    emit statusUpdate(tr("Melde %1 an …").arg(m_db->playersString(Database::Numerus::singular)));
    QApplication::setOverrideCursor(Qt::WaitCursor);

    m_db->registerPlayers(name);
    m_playersName->clear();
    m_playersName->setFocus();
    reload();

    // Make sure the entered name is visible
    int targetRow = 0;
    for (int i = 0; i < m_playersList->rowCount(); i++) {
        if (m_playersList->item(i, 0)->text() == name) {
            targetRow = i;
            break;
        }
    }
    if (targetRow == m_playersList->rowCount() - 1) {
        m_playersList->scrollToBottom();
    } else {
        m_playersList->scrollToItem(m_playersList->item(targetRow, 0),
                                    QAbstractItemView::PositionAtCenter);
    }

    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("%1 „%2“ angemeldet").arg(
        m_db->playersString(Database::Numerus::singular), name));
}

void PlayersPage::editPlayers()
{
    cacheOriginalName();
    m_playersList->editItem(m_playersList->currentItem());
}

void PlayersPage::cacheOriginalName()
{
    m_originalPlayersName = m_playersList->currentItem()->text();
}

void PlayersPage::editEntered()
{
    // Checks a changed player/pair name and saves it if everything is okay.

    QString name = m_playersList->currentItem()->text();

    if (name.isEmpty()) {
        setPlayersName(m_originalPlayersName);
        return;
    }

    if (! m_db->checkName(name)) {
        QMessageBox::critical(this,
            tr("%1 umbenennen nicht möglich").arg(m_db->playersString(Database::Numerus::singular)),
            tr("Der Name „%1“ ist bereits vergeben!").arg(name),
            QMessageBox::Ok);
        setPlayersName(m_originalPlayersName);
        return;
    }

    emit statusUpdate(tr("Benenne %1 um …").arg(m_db->playersString(Database::Numerus::singular)));
    QApplication::setOverrideCursor(Qt::WaitCursor);

    m_db->renamePlayers(m_originalPlayersName, name);
    setPlayersName(name);

    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("%1 „%2“ umbenannt zu „%3“").arg(
        m_db->playersString(Database::Numerus::singular), m_originalPlayersName, name));

    emit playersListChanged();
    reload();
}

void PlayersPage::setPlayersName(QString name)
{
    // Updates the displayed player/pair's name.
    m_playersList->blockSignals(true);
    m_playersList->currentItem()->setText(name);
    m_playersList->blockSignals(false);
}

void PlayersPage::showContextMenu(QPoint point)
{
    QTableWidgetItem *selectedPlayers = m_playersList->currentItem();
    if (selectedPlayers == nullptr) {
        return;
    }

    m_contextMenu->clear();
    m_menuTitle->setText(tr("%1 „%2“").arg(m_db->playersString(Database::Numerus::singular),
                                           selectedPlayers->text()));
    m_contextMenu->addAction(m_menuTitleAction);
    m_contextMenu->addSeparator();
    m_contextMenu->addAction(m_editAction);
    m_contextMenu->addAction(m_deleteAction);

    m_deleteAction->setEnabled(! m_db->tournamentStarted());
    m_contextMenu->exec(m_playersList->mapToGlobal(point));
}

void PlayersPage::deletePlayers()
{
    QString toDelete = m_playersList->currentItem()->text();

    if (QMessageBox::warning(this,
            tr("%1 löschen").arg(m_db->playersString(Database::Numerus::singular)),
            tr("Soll %1 „%2“ wirklich gelöscht werden?").arg(
                m_db->playersString(Database::Numerus::singular, true), toDelete),
        QMessageBox::Yes | QMessageBox::No)
        != QMessageBox::Yes) {
        return;
    }

    m_db->deletePlayers(toDelete);
    reload();
}
