/* Copyright (c) 2012-2017 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "SettingsPage.h"
#include "Database.h"

// Qt includes
#include <QDebug>
#include <QVBoxLayout>
#include <QGroupBox>
#include <QGridLayout>
#include <QLabel>
#include <QRadioButton>
#include <QSpinBox>
#include <QPushButton>
#include <QMessageBox>
#include <QApplication>
#include <QDir>

SettingsPage::SettingsPage(QWidget *parent, Database *db)
    : QWidget(parent), m_db(db)
{
    m_fixedPairsDescription = tr(
        "<p>Die Spieler in einem Paar ändern sich im Verlauf des Turniers nicht.</p>"
        "<p>"
        "Wenn vor dem Turnierbeginn für jedes Paar ein Tisch und die Rolle „Paar 1“ und „Paar 2“ "
        "ausgelost werden, Paar 1 immer sitzen bleibt und Paar 2 nach jeder Runde einen Tisch "
        "weiterrückt, können die passenden Paare und Tische nach der ersten Runde automatisch "
        "gesetzt werden."
        "</p>"
        "<p>"
        "Alternativ kann auch pro Runde ausgelost werden, welche Paare gegeneinander spielen."
        "</p>"
    );

    m_singlePlayersDescription = tr(
        "<p>"
        "Es werden einzelne Spieler angemeldet. "
        "In jeder Runde werden die Paare und Tische neu ausgelost."
        "</p>"
    );

    QVBoxLayout *layout = new QVBoxLayout(this);

    // Database

    m_databaseBox = new QGroupBox(tr("Datenbank"));
    QGridLayout *databaseBoxLayout = new QGridLayout(m_databaseBox);
    layout->addWidget(m_databaseBox);

    QLabel *dbFileHeaderLabel = new QLabel(tr("Turnierdatenbank:"));
    dbFileHeaderLabel->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    databaseBoxLayout->addWidget(dbFileHeaderLabel, 0, 0);

    m_dbFileLabel = new QLabel;
    m_dbFileLabel->setTextInteractionFlags(Qt::TextSelectableByMouse);
    databaseBoxLayout->addWidget(m_dbFileLabel, 0, 1);

    m_mtVersionHeaderLabel = new QLabel(tr("Erstellt mit:"));
    m_mtVersionHeaderLabel->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    databaseBoxLayout->addWidget(m_mtVersionHeaderLabel, 1, 0);
    m_mtVersionLabel = new QLabel;
    m_mtVersionLabel->setTextInteractionFlags(Qt::TextSelectableByMouse);
    databaseBoxLayout->addWidget(m_mtVersionLabel, 1, 1);

    m_databaseBox->hide();

    // Tournament type

    QGroupBox *tournamentTypeBox = new QGroupBox(tr("Turniertyp"));
    QVBoxLayout *tournamentTypeLayout = new QVBoxLayout(tournamentTypeBox);
    layout->addWidget(tournamentTypeBox);

    m_pairs = new QRadioButton(tr("Feste Paare"));
    m_pairs->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    tournamentTypeLayout->addWidget(m_pairs);
    connect(m_pairs, &QRadioButton::clicked,
            this, &SettingsPage::updateTournamentModeDescription);

    m_singlePlayers = new QRadioButton(tr("Einzelne Spieler"));
    m_singlePlayers->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    tournamentTypeLayout->addWidget(m_singlePlayers);
    connect(m_singlePlayers, &QRadioButton::clicked,
            this, &SettingsPage::updateTournamentModeDescription);

    m_tournamentModeHeaderLabel = new QLabel;
    m_tournamentModeHeaderLabel->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    m_tournamentModeHeaderLabel->setStyleSheet(QString::fromUtf8(
        "QLabel { text-decoration: underline; }"));
    tournamentTypeLayout->addWidget(m_tournamentModeHeaderLabel);

    // Adds additional space
    tournamentTypeLayout->addWidget(new QLabel);

    m_tournamentModeLabel = new QLabel;
    m_tournamentModeLabel->setText(m_fixedPairsDescription);
    m_tournamentModeLabel->setWordWrap(true);
    m_tournamentModeLabel->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Minimum);
    tournamentTypeLayout->addWidget(m_tournamentModeLabel);

    // Game settings

    QGroupBox *gameSettingsBox = new QGroupBox(tr("Spieleinstellungen"));
    QHBoxLayout *gameSettingsMainLayout = new QHBoxLayout(gameSettingsBox);
    QGridLayout *gameSettingsLayout = new QGridLayout;
    gameSettingsMainLayout->addLayout(gameSettingsLayout);
    layout->addWidget(gameSettingsBox);

    gameSettingsLayout->addWidget(new QLabel(tr("Punkte pro Bobbl:")), 0, 0);

    m_boogerScore = new QSpinBox;
    m_boogerScore->setMinimum(1);
    m_boogerScore->setMaximum(999);
    gameSettingsLayout->addWidget(m_boogerScore, 0, 1);

    m_boogerScoreLabel = new QLabel;
    gameSettingsLayout->addWidget(m_boogerScoreLabel, 0, 2);

    gameSettingsLayout->addWidget(new QLabel(tr("Bobbl pro Runde:")), 1, 0);

    m_gamesPerRound = new QSpinBox;
    m_gamesPerRound->setMinimum(1);
    m_gamesPerRound->setMaximum(999);
    gameSettingsLayout->addWidget(m_gamesPerRound, 1, 1);

    m_gamesPerRoundLabel = new QLabel;
    gameSettingsLayout->addWidget(m_gamesPerRoundLabel, 1, 2);

    gameSettingsMainLayout->addStretch();

    // Save tournament settings

    QWidget *saveSettingsWidget = new QWidget;
    layout->addWidget(saveSettingsWidget);
    QHBoxLayout *saveSettingsLayout = new QHBoxLayout(saveSettingsWidget);

    m_resetDatabase = new QPushButton(tr("Datenbank zurücksetzen"));
    m_resetDatabase->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    m_resetDatabase->setVisible(false);
    saveSettingsLayout->addWidget(m_resetDatabase);
    connect(m_resetDatabase, &QPushButton::clicked, this, &SettingsPage::resetDatabase);

    m_saveSettings = new QPushButton(tr("Turniereinstellungen speichern"));
    m_saveSettings->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    saveSettingsLayout->addWidget(m_saveSettings);
    connect(m_saveSettings, &QPushButton::clicked, this, &SettingsPage::saveSettings);

    saveSettingsLayout->addStretch();

    layout->addStretch();

    resetGui();
}

void SettingsPage::resetGui()
{
    m_resetDatabase->hide();
    m_saveSettings->show();

    m_tournamentModeHeaderLabel->setText(tr("Feste Paare:"));
    m_tournamentModeHeaderLabel->hide();
    m_tournamentModeLabel->setText(m_fixedPairsDescription);
    m_pairs->show();
    m_singlePlayers->show();

    m_boogerScoreLabel->hide();
    m_boogerScore->show();
    m_gamesPerRoundLabel->hide();
    m_gamesPerRound->show();

    m_databaseBox->hide();
}

void SettingsPage::resetDatabase()
{
    if (QMessageBox::warning(this,
        tr("Datenbank zurücksetzen"),
        tr("Sollen wirklich alle Einstellungen und Ergebnisse gelöscht werden?"),
        QMessageBox::Yes | QMessageBox::No)
        != QMessageBox::Yes) {
        return;
    }

    emit statusUpdate(tr("Setzte Turnierdatenbank zurück …"));
    QApplication::setOverrideCursor(Qt::WaitCursor);

    m_db->reset();
    resetGui();
    reload();
    emit dbReset();

    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Turnierdatenbank zurückgesetzt"));
}

void SettingsPage::saveSettings()
{
    emit statusUpdate(tr("Speichere Turniereinstellungen …"));
    QApplication::setOverrideCursor(Qt::WaitCursor);

    if (m_pairs->isChecked()) {
        m_db->setTournamentMode(Database::TournamentMode::fixedPairs);
    } else {
        m_db->setTournamentMode(Database::TournamentMode::singlePlayers);
    }
    m_db->setBoogerScore(m_boogerScore->value());
    m_db->setGamesPerRound(m_gamesPerRound->value());

    m_db->saveSettings();
    setToSaved();
    emit settingsSaved();
    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Turniereinstellungen gespeichert"));
}

void SettingsPage::updateTournamentModeDescription()
{
    // Display the correct description for the tournament mode.
    if (m_pairs->isChecked()) {
        m_tournamentModeLabel->setText(m_fixedPairsDescription);
        m_tournamentModeHeaderLabel->setText(tr("Feste Paare:"));
    } else {
        m_tournamentModeLabel->setText(m_singlePlayersDescription);
        m_tournamentModeHeaderLabel->setText(tr("Einzelne Spieler:"));
    }
}

void SettingsPage::reload()
{
    // Sets all displayed values accoding to their db values.

    switch (m_db->getTournamentMode()) {
    case Database::TournamentMode::fixedPairs:
        m_pairs->setChecked(Qt::Checked);
        m_singlePlayers->setChecked(Qt::Unchecked);
        break;
    case Database::TournamentMode::singlePlayers:
        m_pairs->setChecked(Qt::Unchecked);
        m_singlePlayers->setChecked(Qt::Checked);
        break;
    }

    updateTournamentModeDescription();
    m_boogerScore->setValue(m_db->getBoogerScore());
    m_gamesPerRound->setValue(m_db->getGamesPerRound());
}

void SettingsPage::setToSaved()
{
    m_resetDatabase->show();
    m_saveSettings->hide();

    m_boogerScoreLabel->setText(QString::number(m_boogerScore->value()));
    m_boogerScoreLabel->show();
    m_boogerScore->hide();
    m_gamesPerRoundLabel->setText(QString::number(m_gamesPerRound->value()));
    m_gamesPerRoundLabel->show();
    m_gamesPerRound->hide();

    m_tournamentModeHeaderLabel->show();
    m_pairs->hide();
    m_singlePlayers->hide();

    m_databaseBox->show();
    m_dbFileLabel->setText(QDir::toNativeSeparators(m_db->getDbFile()));
    m_mtVersionLabel->setText(tr("Muckturnier %1").arg(m_db->getMuckturnierVersion()));
}
